(function(
  // Reliable reference to the global object (i.e. window in browsers).
  global,

  // Dummy constructor that we use as the .constructor property for
  // functions that return Generator objects.
  GeneratorFunction
) {
  var hasOwn = Object.prototype.hasOwnProperty;

  if (global.wrapGenerator) {
    return;
  }

  function wrapGenerator(innerFn, self) {
    return new Generator(innerFn, self || null);
  }

  global.wrapGenerator = wrapGenerator;
  if (typeof exports !== "undefined") {
    exports.wrapGenerator = wrapGenerator;
  }

  var GenStateSuspendedStart = "suspendedStart";
  var GenStateSuspendedYield = "suspendedYield";
  var GenStateExecuting = "executing";
  var GenStateCompleted = "completed";

  // Returning this object from the innerFn has the same effect as
  // breaking out of the dispatch switch statement.
  var ContinueSentinel = {};

  wrapGenerator.mark = function(genFun) {
    genFun.constructor = GeneratorFunction;
    return genFun;
  };

  // Ensure isGeneratorFunction works when Function#name not supported.
  if (GeneratorFunction.name !== "GeneratorFunction") {
    GeneratorFunction.name = "GeneratorFunction";
  }

  wrapGenerator.isGeneratorFunction = function(genFun) {
    var ctor = genFun && genFun.constructor;
    return ctor ? GeneratorFunction.name === ctor.name : false;
  };

  function Generator(innerFn, self) {
    var generator = this;
    var context = new Context();
    var state = GenStateSuspendedStart;

    function invoke(method, arg) {
      if (state === GenStateExecuting) {
        throw new Error("Generator is already running");
      }

      if (state === GenStateCompleted) {
        throw new Error("Generator has already finished");
      }

      while (true) {
        var delegate = context.delegate;
        if (delegate) {
          try {
            var info = delegate.generator[method](arg);

            // Delegate generator ran and handled its own exceptions so
            // regardless of what the method was, we continue as if it is
            // "next" with an undefined arg.
            method = "next";
            arg = void 0;

          } catch (uncaught) {
            context.delegate = null;

            // Like returning generator.throw(uncaught), but without the
            // overhead of an extra function call.
            method = "throw";
            arg = uncaught;

            continue;
          }

          if (info.done) {
            context[delegate.resultName] = info.value;
            context.next = delegate.nextLoc;
          } else {
            state = GenStateSuspendedYield;
            return info;
          }

          context.delegate = null;
        }

        if (method === "next") {
          if (state === GenStateSuspendedStart &&
              typeof arg !== "undefined") {
            // https://people.mozilla.org/~jorendorff/es6-draft.html#sec-generatorresume
            throw new TypeError(
              "attempt to send " + JSON.stringify(arg) + " to newborn generator"
            );
          }

          if (state === GenStateSuspendedYield) {
            context.sent = arg;
          } else {
            delete context.sent;
          }

        } else if (method === "throw") {
          if (state === GenStateSuspendedStart) {
            state = GenStateCompleted;
            throw arg;
          }

          if (context.dispatchException(arg)) {
            // If the dispatched exception was caught by a catch block,
            // then let that catch block handle the exception normally.
            method = "next";
            arg = void 0;
          }
        }

        state = GenStateExecuting;

        try {
          var value = innerFn.call(self, context);

          // If an exception is thrown from innerFn, we leave state ===
          // GenStateExecuting and loop back for another invocation.
          state = context.done
            ? GenStateCompleted
            : GenStateSuspendedYield;

          var info = {
            value: value,
            done: context.done
          };

          if (value === ContinueSentinel) {
            if (context.delegate && method === "next") {
              // Deliberately forget the last sent value so that we don't
              // accidentally pass it on to the delegate.
              arg = void 0;
            }
          } else {
            return info;
          }

        } catch (thrown) {
          if (method === "next") {
            context.dispatchException(thrown);
          } else {
            arg = thrown;
          }
        }
      }
    }

    generator.next = invoke.bind(generator, "next");
    generator.throw = invoke.bind(generator, "throw");
  }

  Generator.prototype.toString = function() {
    return "[object Generator]";
  };

  function Context() {
    this.reset();
  }

  Context.prototype = {
    constructor: Context,

    reset: function() {
      this.next = 0;
      this.sent = void 0;
      this.tryStack = [];
      this.done = false;
      this.delegate = null;

      // Pre-initialize at least 20 temporary variables to enable hidden
      // class optimizations for simple generators.
      for (var tempIndex = 0, tempName;
           hasOwn.call(this, tempName = "t" + tempIndex) || tempIndex < 20;
           ++tempIndex) {
        this[tempName] = null;
      }
    },

    stop: function() {
      this.done = true;

      if (hasOwn.call(this, "thrown")) {
        var thrown = this.thrown;
        delete this.thrown;
        throw thrown;
      }

      return this.rval;
    },

    keys: function(object) {
      return Object.keys(object).reverse();
    },

    pushTry: function(catchLoc, finallyLoc, finallyTempVar) {
      if (finallyLoc) {
        this.tryStack.push({
          finallyLoc: finallyLoc,
          finallyTempVar: finallyTempVar
        });
      }

      if (catchLoc) {
        this.tryStack.push({
          catchLoc: catchLoc
        });
      }
    },

    popCatch: function(catchLoc) {
      var lastIndex = this.tryStack.length - 1;
      var entry = this.tryStack[lastIndex];

      if (entry && entry.catchLoc === catchLoc) {
        this.tryStack.length = lastIndex;
      }
    },

    popFinally: function(finallyLoc) {
      var lastIndex = this.tryStack.length - 1;
      var entry = this.tryStack[lastIndex];

      if (!entry || !hasOwn.call(entry, "finallyLoc")) {
        entry = this.tryStack[--lastIndex];
      }

      if (entry && entry.finallyLoc === finallyLoc) {
        this.tryStack.length = lastIndex;
      }
    },

    dispatchException: function(exception) {
      var finallyEntries = [];
      var caught = false;

      if (this.done) {
        throw exception;
      }

      // Dispatch the exception to the "end" location by default.
      this.thrown = exception;
      this.next = "end";

      for (var i = this.tryStack.length - 1; i >= 0; --i) {
        var entry = this.tryStack[i];
        if (entry.catchLoc) {
          this.next = entry.catchLoc;
          caught = true;
          break;
        } else if (entry.finallyLoc) {
          finallyEntries.push(entry);
        }
      }

      while ((entry = finallyEntries.pop())) {
        this[entry.finallyTempVar] = this.next;
        this.next = entry.finallyLoc;
      }

      return caught;
    },

    delegateYield: function(generator, resultName, nextLoc) {
      this.delegate = {
        generator: generator,
        resultName: resultName,
        nextLoc: nextLoc
      };

      return ContinueSentinel;
    }
  };
}).apply(this, Function("return [this, function GeneratorFunction(){}]")());

var EventEmitter = require('events').EventEmitter;
var inherits = require('util').inherits;
inherits(Channel, EventEmitter);
module.exports = Channel;

function Channel(options) {
  if (!(this instanceof Channel))
    return new Channel(options)

  options = options || {}

  EventEmitter.call(this, options)

  this.concurrency = options.concurrency || Infinity
  // closed by default
  this.closed = !(options.closed === false || options.open)
  this.discard = options.discard

  this.fns = []
  this.results = []

  this.fnIndex = 0
  this.resultIndex = 0
  this.pending = 0
  this.errors = 0

  this.reading = true

  this.next = this.next.bind(this)
}

Object.defineProperty(Channel.prototype, 'pushable', {
  get: function () {
    return this.reading && this.pending < this.concurrency
  }
})

// read queue;

// read queue
Object.defineProperty(Channel.prototype, 'queue', {
  get: function () {
    var queue = this.results.length + this.fns.length
    if (this.discard) queue += this.pending
    return queue
  }
})

// you can read from a channel if there's a read queue or if this channel is not closed;

// you can read from a channel if there's a read queue or if this channel is not closed
Object.defineProperty(Channel.prototype, 'readable', {
  get: function () {
    return this.queue || !this.closed
  }
});

Channel.prototype.open = function () {
  this.closed = false
  this.emit('open')
  return this
}

Channel.prototype.close = function () {
  this.closed = true
  this.emit('close')
  return this
}

// when the channel is open,
// wait for the first push event (returns true)
// or close event (returns false);

// or close event (returns false)
Channel.prototype.pushed = wrapGenerator.mark(function() {
  var self;

  return wrapGenerator(function($ctx0) {
    while (1) switch ($ctx0.next) {
    case 0:
      if (!this.closed) {
        $ctx0.next = 5;
        break;
      }

      $ctx0.rval = false;
      delete $ctx0.thrown;
      $ctx0.next = 17;
      break;
    case 5:
      if (!this.queue) {
        $ctx0.next = 10;
        break;
      }

      $ctx0.rval = true;
      delete $ctx0.thrown;
      $ctx0.next = 17;
      break;
    case 10:
      self = this;
      $ctx0.next = 13;

      return function (done) {
        self.on('close', close)
        self.on('push', push)

        function close() {
          cleanup()
          done(null, false)
        }

        function push() {
          cleanup()
          done(null, true)
        }

        function cleanup() {
          self.removeListener('close', close)
          self.removeListener('push', push)
        }
      }
    case 13:
      $ctx0.rval = $ctx0.sent;
      delete $ctx0.thrown;
      $ctx0.next = 17;
      break;
    case 17:
    case "end":
      return $ctx0.stop();
    }
  }, this);
})

/**
 * Push a function to the channel.
 * If `null`, just means closing the channel.
 */;

/**
 * Push a function to the channel.
 * If `null`, just means closing the channel.
 */

Channel.prototype.push = function (fn) {
  if (fn == null) return this.close()

  if (typeof fn !== 'function')
    throw new TypeError('you may only push functions')

  this.fns.push(fn)
  if (!this.discard) this.results.length++
  this.call()
  this.emit('push')
  return this
}

Channel.prototype.call = function () {
  if (!this.pushable) return
  if (!this.fns.length) return

  var fn = this.fns.shift()
  var index = this.fnIndex++
  this.pending++

  var self = this
  fn(function (err, res) {
    self.pending--
    if (err) {
      self.reading = false
      self.errors++
      if (self.discard) {
        self.results.push(err)
      } else {
        self.results[index - self.resultIndex] = err
      }
    } else if (!self.discard) {
      self.results[index - self.resultIndex] = arguments.length > 2
        ? [].slice.call(arguments, 1)
        : res
    }

    self.emit(String(index))
    self.emit('callback')
    self.call()
  })
}

Channel.prototype.read = wrapGenerator.mark(function() {
  var self, res;

  return wrapGenerator(function($ctx1) {
    while (1) switch ($ctx1.next) {
    case 0:
      self = this;

      // continue executing callbacks if no errors occured
      if (!this.reading && !this.errors) this.reading = true;

      if (!!self.discard) {
        $ctx1.next = 15;
        break;
      }

      if (!(self.results.length && 0 in self.results)) {
        $ctx1.next = 13;
        break;
      }

      res = self.results.shift();
      self.resultIndex++;

      if (!(res instanceof Error)) {
        $ctx1.next = 9;
        break;
      }

      self.errors--;
      throw res;
    case 9:
      $ctx1.rval = res;
      delete $ctx1.thrown;
      $ctx1.next = 24;
      break;
    case 13:
      $ctx1.next = 18;
      break;
    case 15:
      if (!self.results.length) {
        $ctx1.next = 18;
        break;
      }

      // these can only be errors
      self.errors--;

      throw self.results.shift();
    case 18:
      $ctx1.next = 20;
      return this.next;
    case 20:
      $ctx1.rval = $ctx1.sent;
      delete $ctx1.thrown;
      $ctx1.next = 24;
      break;
    case 24:
    case "end":
      return $ctx1.stop();
    }
  }, this);
});

Channel.prototype.next = function (done) {
  var event = this.discard
    ? 'callback'
    : String(this.resultIndex)

  var self = this
  this.once(event, onevent)
  if (!this.closed && !this.queue) this.once('close', onclose)

  function onevent() {
    cleanup()
    if (!this.discard) {
      var res = this.results.shift()
      this.resultIndex++
      if (res instanceof Error) {
        this.errors--
        done(res)
      } else {
        done(null, res)
      }
    } else if (this.results.length) {
      // these can only be errors
      this.errors--
      done(this.results.shift())
    } else {
      done()
    }
  }

  function onclose() {
    cleanup()
    done()
  }

  function cleanup() {
    self.removeListener(event, onevent)
    self.removeListener('close', onclose)
  }
}

Channel.prototype.flush = wrapGenerator.mark(function() {
  var results;

  return wrapGenerator(function($ctx2) {
    while (1) switch ($ctx2.next) {
    case 0:
      if (!this.discard) {
        $ctx2.next = 8;
        break;
      }
    case 1:
      if (!this.readable) {
        $ctx2.next = 5;
        break;
      }

      return $ctx2.delegateYield(this.read(), "t0", 3);
    case 3:
      $ctx2.next = 1;
      break;
    case 5:
      delete $ctx2.thrown;
      $ctx2.next = 18;
      break;
    case 8:
      results = [];
    case 9:
      if (!this.readable) {
        $ctx2.next = 14;
        break;
      }

      return $ctx2.delegateYield(this.read(), "t1", 11);
    case 11:
      results.push($ctx2.t1);
      $ctx2.next = 9;
      break;
    case 14:
      $ctx2.rval = results;
      delete $ctx2.thrown;
      $ctx2.next = 18;
      break;
    case 18:
    case "end":
      return $ctx2.stop();
    }
  }, this);
});